"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.qTree = void 0;
const bounds_1 = require("./bounds");
class qTree {
    constructor([x0, y0, x1, y1], options = {}, boundsCache, quadCache, depth = 0) {
        const { maxDepth = 7, maxChildren = 10 } = options;
        this.maxDepth = maxDepth;
        this.maxChildren = maxChildren;
        this.boundsCache = boundsCache;
        this.quadCache = quadCache;
        this.depth = depth;
        const midX = (x0 + x1) / 2;
        const midY = (y0 + y1) / 2;
        this.childAreas = [
            [x0, y0, midX, midY],
            [midX, y0, x1, midY],
            [x0, midY, midX, y1],
            [midX, midY, x1, y1],
        ];
        this.isSplit = false;
        this.children = [null, null, null, null];
        this.items = new Set();
    }
    clear() {
        this.items.clear();
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            child.clear();
            this.children[i] = null;
        }
    }
    insert(item) {
        if (!this.isSplit && this.items.size + 1 > this.maxChildren) {
            this.isSplit = true;
            for (const i of this.items)
                if (this.tryChildInsert(i))
                    this.items.delete(i);
        }
        if (this.isSplit && this.depth + 1 < this.maxDepth) {
            if (this.tryChildInsert(item))
                return;
        }
        this.items.add(item);
        this.quadCache.set(item, this);
    }
    tryChildInsert(item) {
        for (let i = 0; i < 4; i++) {
            const childRect = this.childAreas[i];
            const rect = this.boundsCache.get(item);
            if ((0, bounds_1.contains)(childRect, rect)) {
                let child = this.children[i];
                if (!child) {
                    this.children[i] = child = new qTree(childRect, { maxDepth: this.maxDepth }, this.boundsCache, this.quadCache, this.depth + 1);
                }
                child.insert(item);
                return true;
            }
        }
        return false;
    }
    remove(item) {
        if (this.items.has(item)) {
            this.items.delete(item);
            this.quadCache.delete(item);
            this.boundsCache.delete(item);
            // TODO handle cleanup
            return true;
        }
        const quad = this.quadCache.get(item);
        if (!quad)
            return false;
        return quad.remove(item);
    }
    collectAll(result = []) {
        for (const r of this.items)
            result.push(r);
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            child.collectAll(result);
        }
        return result;
    }
    collect(area, result = []) {
        for (const item of this.items) {
            const rect = this.boundsCache.get(item);
            if ((0, bounds_1.overlaps)(area, rect))
                result.push(item);
        }
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            const childArea = this.childAreas[i];
            if ((0, bounds_1.contains)(area, childArea))
                child.collectAll(result);
            else if ((0, bounds_1.overlaps)(area, childArea))
                child.collect(area, result);
        }
        return result;
    }
    *all() {
        yield* this.items;
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            yield* child.all();
        }
    }
    *search(area) {
        for (const item of this.items) {
            const rect = this.boundsCache.get(item);
            if ((0, bounds_1.overlaps)(area, rect))
                yield item;
        }
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            const childArea = this.childAreas[i];
            if ((0, bounds_1.contains)(area, childArea))
                yield* child.all();
            else if ((0, bounds_1.overlaps)(area, childArea))
                yield* child.search(area);
        }
    }
    size() {
        let count = this.items.size;
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            count += child.size();
        }
        return count;
    }
}
exports.qTree = qTree;
//# sourceMappingURL=q-tree.js.map